/* Copyright(c) 2002,2003,2004 Device Drivers Limited, All rights reserved. */
/*                              info@devdrv.co.jp http://www.devdrv.co.jp/ */
/*
 * usage:
 *   insmod loop.ko            -- default, use atomic function
 *   insmod loop.ko debug=1    -- use spinlock
 *   insmod loop.ko debug=2    -- use atomic function and schedule()
 *   insmod loop.ko debug=3    -- use spinlock and schedule()
 *   insmod loop.ko debug=4... -- with debug meddage(not worked well)
 *   insmod loop.ko debug=8... -- with verbose meddage(not worked well)
 *
 *   # echo 5000 > /proc/loop
 *   # cat /proc/loop
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>   /* printk() */
#include <linux/fs.h>       /* everything... */
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/spinlock.h>
#include <linux/random.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>

#define BUFSIZ 128
static char st_b[BUFSIZ] = "0\n"; /* data buffer, initialize value is "0" */

#define ORIGINAL_LOGIC 1

#define IS_SPINLOCK(deb) (deb & 0x01)
#define IS_SCHEDULE(deb) (deb & 0x02)
#define IS_MESSAGE(deb)  (deb & 0x04)
#define IS_VERBOSE(deb)  (deb & 0x08)
#define IS_DEBUG(deb)    (deb & 0x10)

static int debug;
static atomic_t sync;
static spinlock_t slock = SPIN_LOCK_UNLOCKED;

struct fortune {
	atomic_t lock;
	unsigned long data;
	unsigned char cpu;
} table[256];

static struct wq {
	int num;
	char *name;
} loop_data;

static struct workqueue_struct *loop;
struct work_struct work;

static void looping(void *data)
{
	unsigned long ms;
	unsigned long junk;
	struct fortune *p;
	unsigned long i, j = 0;
	int val, old, new;

	/* spin_lock(&slock); */
	old = atomic_read(&sync);
	atomic_inc(&sync);
	new = atomic_read(&sync);
	/* spin_unlock(&slock); */

	i = 0;
	while((val = atomic_read(&sync)) <= 1) {
		schedule();
		spin_lock(&slock);
		if (i == 100000 || i == 0) {
			printk("** (%d) ** val = %d (old = %d, new = %d)\n",
			       smp_processor_id(), val, old, new);
			i = 0;
		}
		spin_unlock(&slock);
		i++;
	}
	if(IS_MESSAGE(debug)) {
		printk("** OK ** (%d) ** val = %d, i = %lu\n",
		       smp_processor_id(), val, i);
	}
	ms = (unsigned long) ((struct wq *)data)->num;
	if (ms > 0x1000000) ms = 0x1000000;
	if(IS_VERBOSE(debug)) {
		printk("**(%d) looping start[%s] cnt = %lu\n",
		       smp_processor_id(), current->comm, ms);
		j = jiffies;
	}
	for(i = 0; i < ms; i++) {

		if (IS_SCHEDULE(debug))
			schedule();

		get_random_bytes(&junk, sizeof(junk));
		/* random counter up */
		junk &= 0xFF;

		p = &table[junk];

		if (IS_SPINLOCK(debug)) {
			spin_lock(&slock);
			if (p->cpu == 0) {
				p->cpu = smp_processor_id() + 1;
			}
			if (p->cpu == (smp_processor_id() + 1)) {
				p->data++;
			}
			spin_unlock(&slock);
		} else { /* atomic */
			if (atomic_dec_and_test(&p->lock) == 1) {
				p->cpu = smp_processor_id() + 1;
			}
			else {
				atomic_inc(&p->lock);
			}
			if (p->cpu == (smp_processor_id() + 1)) {
				p->data++;
			}
		}
		if (IS_DEBUG(debug) && (junk != 0 && (jiffies & 0xFF)))
			printk("**junk = %lu\n", junk);
	}
	if(IS_VERBOSE(debug)) {
		printk("**(%d) looping end[%s] = %lu\n",
		       smp_processor_id(), current->comm, jiffies - j);
	}
	atomic_dec(&sync);
	/* return jiffies; */
}

static inline void wake(void)
{
        queue_work(loop, &work);
}

/*
 * loop_read_proc -- called when user is reading
 */
static
int loop_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
    int len, i, j;
    struct fortune *p;
    unsigned long temp;

    if (IS_VERBOSE(debug)) {
	    printk("read_proc(), count = %d, off = %d\n",
		   count, (int) offset);
    }
    flush_workqueue(loop);

    p = &table[0];
    len = sprintf(buf,
    "     -0   -1   -2   -3   -4   -5   -6   -7   -8   -9   -A   -B   -C   -D   -E   -F\n");
    for(i = 0; i < 16; i++) {
	    len += sprintf(buf + len, "%01X-", i);
	    for(j = 0; j < 16; j++) {
		    temp = p->cpu == 1 ? p->data : 0;
		    if (temp > 0xFFFF)
			    temp = 0xFFFF;
		    len += sprintf(buf + len, " %04lx", temp);
		    p++;
	    }
	    len += sprintf(buf + len, "\n");
    }
    p = &table[0];
    len += sprintf(buf + len, "\n");
    for(i = 0; i < 16; i++) {
	    len += sprintf(buf + len, "%01X-", i);
	    for(j = 0; j < 16; j++) {
		    temp = p->cpu == 2 ? p->data : 0;
		    if (temp > 0xFFFF)
			    temp = 0xFFFF;
		    len += sprintf(buf + len, " %04lx", temp);
		    p++;
	    }
	    len += sprintf(buf + len, "\n");
    }
    *eof = 1;
    return len;
}

/*
 * loop_write_proc -- called when user is writing
 */
int loop_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
    int length, lcnt, i;
    struct fortune *p;
    struct wq *datap = &loop_data;

    length = count > (BUFSIZ - 1) ? (BUFSIZ - 1) : count;
    lcnt = simple_strtoul(buf, NULL, 10);

    /* initialize table */
    p = &table[0];
    for(i = 0; i < 256; i++) {
	    atomic_set(&p->lock, 1);
	    p->data = 0;
	    p->cpu = 0;
	    p++;
    }
    atomic_set(&sync, 0);

    /* start... */
    datap->num = lcnt;
    wake();
    wake();

    return length;
}

static void loop_create_proc(void)
{
    struct proc_dir_entry *entry;
    struct fortune *p;
    int i;

    p = &table[0];
    for(i = 0; i < 256; i++) {
	    atomic_set(&p->lock, 1);
	    p->data = 0;
	    p->cpu = 0;
	    p++;
    }
    atomic_set(&sync, 0);

    entry = create_proc_entry("loop", 0, 0); /* register "loop" procfile */
    entry->read_proc = loop_read_proc; /* register read routine*/
    entry->write_proc = loop_write_proc; /* register write routine */
}

static void loop_remove_proc(void)
{
    remove_proc_entry("loop", NULL);
}

static int __init loop_init_module(void)
{
	loop_create_proc();
	loop = create_workqueue("loopd");
	if (!loop)
		panic("Failed to create loopd\n");
	INIT_WORK(&work, looping, &loop_data);

	if (IS_VERBOSE(debug))
		printk("loop, world data = %s\n", st_b);
	return 0;
}

static void __exit loop_cleanup_module(void)
{
	if (IS_VERBOSE(debug))
		printk("Goodbye loop data = %s\n", st_b);
	destroy_workqueue(loop);
	loop_remove_proc();
}

MODULE_DESCRIPTION("Sample Loop Module");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Atomu Hidaka <info@devdrv.co.jp>");
MODULE_PARM(debug,       "i");
MODULE_PARM_DESC(debug, "loop debug level (0-15)");

module_init(loop_init_module);
module_exit(loop_cleanup_module);
