/* Copyright(c) 2002,2003,2004 Device Drivers Limited, All rights reserved. */
/*                              info@devdrv.com http://www.devdrv.com/ */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <asm/atomic.h>
#include <linux/spinlock.h>
#include <linux/random.h>
#include <asm/uaccess.h>

#define BUFSIZ 128
/* #define ORIGINAL_LOGIC 1 */

static int debug;
static atomic_t sync;
static spinlock_t slock = SPIN_LOCK_UNLOCKED;

struct fortune {
	atomic_t lock;
	unsigned long data;
	unsigned char cpu;
} table[256];

static unsigned int looping(unsigned long ms)
{
	unsigned long junk;
	struct fortune *p;
	unsigned long i;

	for(i = 0; i < ms; i++) {

		if (debug == 1)
			schedule();

		get_random_bytes(&junk, sizeof(junk));
		/* random counter up */
		junk &= 0xFF;

		p = &table[junk];
#ifdef ORIGINAL_LOGIC
		spin_lock(&slock);
		if (p->cpu == 0) {
			p->cpu = smp_processor_id() + 1;
		}
		if (p->cpu == (smp_processor_id() + 1)) {
			p->data++;
		}
		spin_unlock(&slock);
#else
		if (atomic_dec_and_test(&p->lock) == 1) { /* p->lock == 0 */
			p->cpu = smp_processor_id() + 1;
		}
		else {
			atomic_inc(&p->lock);
		}
		if (p->cpu == (smp_processor_id() + 1)) {
			p->data++;
		}
#endif
		if (debug > 4 && (junk != 0 && (jiffies & 0xFF)))
			printk("**junk = %lu\n", junk);
	}
	if(debug > 3) {
		printk("**(%d) %d: looping end = %lu\n",
		       smp_processor_id(), current->pid, jiffies);
	}
	return jiffies;
}

/*
 * loop_read_proc --
 */
static
int loop_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
    int len, i, j;
    struct fortune *p;
    unsigned long temp;

    printk("read_proc(), count = %d, off = %d\n",
	   count, (int) offset);

    p = &table[0];
    len = sprintf(buf,
    "     -0   -1   -2   -3   -4   -5   -6   -7   -8   -9   -A   -B   -C   -D   -E   -F\n");
    for(i = 0; i < 16; i++) {
	    len += sprintf(buf + len, "%01X-", i);
	    for(j = 0; j < 16; j++) {
		    temp = p->cpu == 1 ? p->data : 0;
		    if (temp > 0xFFFF)
			    temp = 0xFFFF;
		    len += sprintf(buf + len, " %04lx", temp);
		    p++;
	    }
	    len += sprintf(buf + len, "\n");
    }
    p = &table[0];
    len += sprintf(buf + len, "\n");
    for(i = 0; i < 16; i++) {
	    len += sprintf(buf + len, "%01X-", i);
	    for(j = 0; j < 16; j++) {
		    temp = p->cpu == 2 ? p->data : 0;
		    if (temp > 0xFFFF)
			    temp = 0xFFFF;
		    len += sprintf(buf + len, " %04lx", temp);
		    p++;
	    }
	    len += sprintf(buf + len, "\n");
    }
    *eof = 1;
    return len;
}

/*
 * loop_write_proc --
 */
static
int loop_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
    int length;
    int loop, i, val, old, new;
    char buffer[BUFSIZ];
    struct fortune *p;

    length = count > (BUFSIZ - 1) ? (BUFSIZ - 1) : count;
	   
    copy_from_user(buffer, buf, length);
    buffer[length] = '\0';

    if (debug > 4) {
	    printk("write_loop(), buf = %s, count = %d, length = %d\n",
		   buffer, (int) count, length);
    }
    loop = simple_strtoul(buffer, NULL, 10);
    if (loop == 0) {
	    /* initialize table */
	    p = &table[0];
	    for(i = 0; i < 256; i++) {
		    atomic_set(&p->lock, 1);
		    p->data = 0;
		    p->cpu = 0;
		    p++;
	    }
	    atomic_set(&sync, 0);
    }
    else {
	    old = atomic_read(&sync);
	    atomic_inc(&sync);
	    new = atomic_read(&sync);

	    i = 0;
	    while((val = atomic_read(&sync)) <= 1) {
		    schedule();
		    spin_lock(&slock);
		    if (i == 100000 || i == 0) {
			    printk("** (%d) ** val = %d (old = %d, new = %d)\n",
				   smp_processor_id(), val, old, new);
			    i = 0;
		    }
		    spin_unlock(&slock);
		    i++;
	    }
	    if (debug > 2) {
		    printk("** OK ** (%d) ** val = %d, i = %d\n",
			   smp_processor_id(), val, i);
	    }
	    /* action ! */
	    (void) looping(loop);

	    atomic_dec(&sync);
    }
    return length;
}

static void loop_create_proc(void)
{
    struct proc_dir_entry *entry;
    struct fortune *p;
    int i;

    p = &table[0];
    for(i = 0; i < 256; i++) {
	    atomic_set(&p->lock, 1);
	    p->data = 0;
	    p->cpu = 0;
	    p++;
    }
    atomic_set(&sync, 0);

    entry = create_proc_entry("loop0", 0, 0);
    entry->read_proc = loop_read_proc;
    entry->write_proc = loop_write_proc;

    entry = create_proc_entry("loop1", 0, 0);
    entry->read_proc = loop_read_proc;
    entry->write_proc = loop_write_proc;
}

static void loop_remove_proc(void)
{
    remove_proc_entry("loop0", NULL);
    remove_proc_entry("loop1", NULL);
}

static int __init loop_init_module(void)
{
    loop_create_proc();
    printk("loop2, init data\n");
    return 0;
}

static void __exit loop_cleanup_module(void)
{
    printk("loop2, exit data\n");
    loop_remove_proc();
}

MODULE_DESCRIPTION("loop2 Module for SMP test");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Atomu Hidaka <info@devdrv.co.jp>");
MODULE_PARM(debug,       "i");
MODULE_PARM_DESC(debug, "loop debug level (0-6)");

module_init(loop_init_module);
module_exit(loop_cleanup_module);
