/* Copyright(c) 2002,2003 Device Drivers Limited, All rights reserved. */
/*                              info@devdrv.com http://www.devdrv.com/ */
#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>   /* printk() */
#include <linux/fs.h>       /* everything... */
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <asm/uaccess.h>

#define BUFSIZ 128
static char st_b[BUFSIZ] = "0\n"; /* data buffer, initialize value is "0" */

static struct wq {
	int num;
	char *name;
} workq_data;

static void deferred_print(void *data)
{
	printk("workq(%d): deferred_print = %s [%s]\n",
	       smp_processor_id(), ((struct wq *)data)->name, current->comm);
}

DECLARE_WORK(workq, deferred_print, (void *) &workq_data); 

/*
 * workq_read_proc -- called when user is reading
 */
int workq_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
    int return_length = 0;
    struct wq *datap = &workq_data;

    printk("read_workq(%d) = %s, count = %d, off = %ld [%s]\n",
	   smp_processor_id(),
	   st_b, count, offset, current->comm);

    if (offset == 0) {
	    return_length = strlen(st_b);
	    strcpy(buf, st_b);

	    datap->name = "READ";
	    schedule_work(&workq);

    }
    *eof = 1;
    *start = buf + offset; /* update for next start point */
    return return_length;
}

/*
 * workq_read_proc -- called when user is writing
 */
int workq_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
    int length;
    struct wq *datap = &workq_data;

    length = count > (BUFSIZ - 1) ? (BUFSIZ - 1) : count; /* max bytes limit */

    copy_from_user(st_b, buf, length);
    st_b[length] = '\0';

    printk("write_workq(%d), buf = %s, count = %lu, length = %d [%s]\n",
	   smp_processor_id(),
	   st_b, count, length, current->comm);

    datap->name = "WRITE";
    schedule_work(&workq);

    return length;
}

static void workq_create_proc(void)
{
    struct proc_dir_entry *entry;
    entry = create_proc_entry("workq", 0, 0); /* register "workq" procfile */
    entry->read_proc = workq_read_proc; /* register read routine*/
    entry->write_proc = workq_write_proc; /* register write routine */
}

static void workq_remove_proc(void)
{
    remove_proc_entry("workq", NULL /* parent dir */);
}

static int __init workq_init_module(void)
{
    workq_create_proc();
    printk("<1>Workq, world data = %s\n", st_b);
    return 0;
}

static void __exit workq_cleanup_module(void)
{
    printk("<1>Goodbye world data = %s\n", st_b);
    workq_remove_proc();
}

MODULE_DESCRIPTION("Sample Workq Module");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Atomu Hidaka <hidaka@devdrv.com>");
MODULE_PARM(debug,       "i");
MODULE_PARM_DESC(debug, "workq debug level (0-6)");

module_init(workq_init_module);
module_exit(workq_cleanup_module);
