/* Copyright(c) 2002,2003 Device Drivers Limited, All rights reserved. */
/*                              info@devdrv.com http://www.devdrv.com/ */
#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>   /* printk() */
#include <linux/fs.h>       /* everything... */
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <asm/uaccess.h>

#define BUFSIZ 128
static char st_b[BUFSIZ] = "0\n"; /* data buffer, initialize value is "0" */

static struct wq {
	int num;
	char *name;
} wq2_data;

static void deferred_print(void *data)
{
	printk("wq2(%d): deferred_print = %s [%s]\n",
	       smp_processor_id(), ((struct wq *)data)->name, current->comm);
}

/* DECLARE_WORK(wq2, deferred_print, (void *) &wq2_data);  */

static struct workqueue_struct *wq2;
struct work_struct work;

static inline void wake(void)
{
        queue_work(wq2, &work);
}

/*
 * wq2_read_proc -- called when user is reading
 */
int wq2_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
    int return_length = 0;
    struct wq *datap = &wq2_data;

    printk("read_wq2(%d) = %s, count = %d, off = %ld [%s]\n",
	   smp_processor_id(),
	   st_b, count, offset, current->comm);

    if (offset == 0) {
	    return_length = strlen(st_b);
	    strcpy(buf, st_b);

	    datap->name = "READ";
	    /* schedule_work(&wq2); */
	    wake();
	    wake();

    }
    *eof = 1;
    *start = buf + offset; /* update for next start point */
    return return_length;
}

/*
 * wq2_read_proc -- called when user is writing
 */
int wq2_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
    int length;
    struct wq *datap = &wq2_data;

    length = count > (BUFSIZ - 1) ? (BUFSIZ - 1) : count; /* max bytes limit */

    copy_from_user(st_b, buf, length);
    st_b[length] = '\0';

    printk("write_wq2(%d), buf = %s, count = %lu, length = %d [%s]\n",
	   smp_processor_id(),
	   st_b, count, length, current->comm);

    datap->name = "WRITE";
    /* schedule_work(&wq2); */
    wake();

    return length;
}

static void wq2_create_proc(void)
{
    struct proc_dir_entry *entry;
    entry = create_proc_entry("wq2", 0, 0); /* register "wq2" procfile */
    entry->read_proc = wq2_read_proc; /* register read routine*/
    entry->write_proc = wq2_write_proc; /* register write routine */
}

static void wq2_remove_proc(void)
{
    remove_proc_entry("wq2", NULL /* parent dir */);
}

static int __init wq2_init_module(void)
{
    wq2_create_proc();

    wq2 = create_workqueue("wq2d");
    if (!wq2)
	    panic("Failed to create wq2d\n");
    INIT_WORK(&work, deferred_print, &wq2_data);

    printk("<1>Wq2, world data = %s\n", st_b);
    return 0;
}

static void __exit wq2_cleanup_module(void)
{
    printk("<1>Goodbye world data = %s\n", st_b);

    destroy_workqueue(wq2);

    wq2_remove_proc();
}

MODULE_DESCRIPTION("Sample Wq2 Module");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Atomu Hidaka <hidaka@devdrv.com>");
MODULE_PARM(debug,       "i");
MODULE_PARM_DESC(debug, "wq2 debug level (0-6)");

module_init(wq2_init_module);
module_exit(wq2_cleanup_module);
